// Source file for mesh class



// Include files

#include "R3Shapes/R3Shapes.h"
#include "mesh.h"



// Class variables

static int next_vertex_id = 0;
static int next_face_id = 0;



////////////////////////////////////////////////////////////
// VERTEX MEMBER FUNCTIONS
////////////////////////////////////////////////////////////

MeshVertex::
MeshVertex(void)
  : position(0, 0, 0),
    id(next_vertex_id++)
{
}



MeshVertex::
MeshVertex(const R3Point& position)
  : position(position),
    id(next_vertex_id++)
{
}



MeshVertex::
MeshVertex(double x, double y, double z)
  : position(x, y, z),
    id(next_vertex_id++)
{
}



////////////////////////////////////////////////////////////
// FACE MEMBER FUNCTIONS
////////////////////////////////////////////////////////////

MeshFace::
MeshFace(void)
  : plane(0, 0, 0, 0),
    id(next_face_id++)
{
}



MeshFace::
MeshFace(MeshVertex *v1, MeshVertex *v2, MeshVertex *v3)
  : plane(0, 0, 0, 0),
    id(next_face_id++)
{
  // Assign vertices
  vertices.Insert(v1);
  vertices.Insert(v2);
  vertices.Insert(v3);

  // Assign plane
  plane = R3Plane(v1->position, v2->position, v3->position);
}



MeshFace::
MeshFace(const RNArray<MeshVertex *>& verts)
  : plane(0, 0, 0, 0),
    id(next_face_id++)
{
  // Assign vertices
  RNArray<R3Point *> points;
  for (int i = 0; i < verts.NEntries(); i++) {
    vertices.Insert(verts[i]);
    points.Insert(&(verts[i]->position));
  }

  // Assign plane
  if (NVertices() > 2) {
    plane = R3Plane(points);
  }
}




////////////////////////////////////////////////////////////
// MESH MEMBER FUNCTIONS
////////////////////////////////////////////////////////////

Mesh::
Mesh(void)
  : bbox(R3null_box)
{
}



////////////////////////////////////////////////////////////
// MESH OFF FILE INPUT/OUTPUT
////////////////////////////////////////////////////////////

int Mesh::
ReadOffFile(const char *filename)
{
  // Open file
  FILE *fp = fopen(filename, "r");
  if (!fp) {
    fprintf(stderr, "Unable to open file %s\n", filename);
    return 0;
  }

  // Read file
  int nverts = 0;
  int nfaces = 0;
  int nedges = 0;
  int line_count = 0;
  char buffer[1024];
  while (fgets(buffer, 1023, fp)) {
    // Increment line counter
    line_count++;

    // Skip white space
    char *bufferp = buffer;
    while (isspace(*bufferp)) bufferp++;

    // Skip blank lines and comments
    if (*bufferp == '#') continue;
    if (*bufferp == '\0') continue;

    // Check section
    if (nverts == 0) {
      // Read header 
      if (!strstr(bufferp, "OFF")) {
        // Read mesh counts
        if ((sscanf(bufferp, "%d%d%d", &nverts, &nfaces, &nedges) != 3) || (nverts == 0)) {
          fprintf(stderr, "Syntax error reading header on line %d in file %s\n", line_count, filename);
          fclose(fp);
          return 0;
        }

        // Check mesh counts
        if ((nverts <= 0) || (nfaces <= 0)) {
          fprintf(stderr, "Invalid number of verts or faces in off file: %s\n", filename);
          return 0;
        }
      }
    }
    else if (NVertices() < nverts) {
      // Read vertex coordinates
      double x, y, z;
      if (sscanf(bufferp, "%lf%lf%lf", &x, &y, &z) != 3) {
        fprintf(stderr, "Syntax error with vertex coordinates on line %d in file %s\n", line_count, filename);
        fclose(fp);
        return 0;
      }

      // Create vertex
      MeshVertex *vertex = new MeshVertex(x, y, z);
      assert(vertex);

      // Add vertex to mesh
      vertices.Insert(vertex);
      bbox.Union(vertex->position);
    }
    else if (NFaces() < nfaces) {
      // Read number of vertices in face 
      int face_nverts = 0;
      bufferp = strtok(bufferp, " \t");
      if (bufferp) face_nverts = atoi(bufferp);
      else {
        fprintf(stderr, "Syntax error with face on line %d in file %s\n", line_count, filename);
        fclose(fp);
        return 0;
      }

      // Read array of vertices for face
      RNArray<MeshVertex *> face_verts;
      for (int i = 0; i < face_nverts; i++) {
        bufferp = strtok(NULL, " \t");
        if (bufferp) {
          int vert_index = atoi(bufferp);
          MeshVertex *vertex = Vertex(vert_index);
          face_verts.Insert(vertex);
        }
        else {
          fprintf(stderr, "Syntax error with face on line %d in file %s\n", line_count, filename);
          fclose(fp);
          return 0;
        }
      }

      // Create face
      MeshFace *face = new MeshFace(face_verts);
      assert(face);

      // Add face to mesh
      faces.Insert(face);
    }
    else {
      // Should never get here
      fprintf(stderr, "Found extra text starting at line %d in file %s\n", line_count, filename);
      break;
    }
  }

  // Check whether read all faces
  if (nfaces != NFaces()) {
    fprintf(stderr, "Expected %d faces, but read only %d faces in file %s\n", nfaces, NFaces(), filename);
  }

  // Close file
  fclose(fp);

  // Return number of faces
  return NFaces();
}



int Mesh::
WriteOffFile(const char *filename)
{
  // Open file
  FILE *fp = fopen(filename, "w");
  if (!fp) {
    fprintf(stderr, "Unable to open file %s\n", filename);
    return 0;
  }

  // Write header
  fprintf(fp, "OFF\n");
  fprintf(fp, "%d %d %d\n", NVertices(), NFaces(), 0);

  // Write vertices
  for (int i = 0; i < NVertices(); i++) {
    MeshVertex *vertex = Vertex(i);
    const R3Point& p = vertex->position;
    fprintf(fp, "%g %g %g\n", p.X(), p.Y(), p.Z());
  }

  // Write Faces
  for (int i = 0; i < NFaces(); i++) {
    MeshFace *face = Face(i);
    fprintf(fp, "%d\n", face->NVertices());
    for (int j = 0; j < face->NVertices(); j++) {
      fprintf(fp, " %d", face->Vertex(j)->id);
    }
    fprintf(fp, "\n");
  }

  // Close file
  fclose(fp);

  // Return number of faces
  return NFaces();
}
